package csci4534.controller;

/**
 * The location information of a byte under a controller, in terms of
 * the controller's disks. In other words, it maps a controller byte
 * offset onto the associated disk, block, and block data offset. It
 * also offers convenience methods to read/write that byte.
 * 
 * @author Toli Lerios
 **/

public class BytePosition
{

    // INSTANCE DATA.

    private Disk[] mDisks;
    private int mBlockOffset;

    /**
     * The index of the disk containing the byte, among the array of
     * the controller's disks. It should be set by a subclass.
     **/

    protected int mDiskIndex;

    /**
     * The index of the block containing the byte, among the blocks of
     * the containing disk. It should be set by a subclass.
     **/

    protected int mBlockIndex;


    // CONSTRUCTORS.

    /**
     * Creates new location information for the given byte offset and
     * the given controller's disks. It only sets the byte's block
     * data offset and retains the controller's disks; subclass
     * constructors should set the remaining location information.
     *
     * @param offset The controller's byte offset.
     * @param disks The controller's disks.
     **/

    protected BytePosition(int offset,
                           Disk[] disks)
    {
        mDisks=disks;
        mBlockOffset=offset%getBlockSize();
    }


    // PUBLIC INTERFACE.

    /**
     * Returns the controller's disks.
     *
     * @return The disks.
     **/

    public Disk[] getDisks()
    {
        return mDisks;
    }

    /**
     * Returns the number <EM>n<SUB>b</SUB></EM> of blocks on each of
     * the controller's disks.
     *
     * @return The number of blocks.
     **/

    public int getBlockCount()
    {
        return getDisks()[0].getBlockCount();
    }

    /**
     * Returns the block size <EM>c<SUB>b</SUB></EM> of the
     * controller's disks.
     *
     * @return The block size, in bytes.
     **/

    public int getBlockSize()
    {
        return getDisks()[0].getBlockSize();
    }

    /**
     * Returns the index of the disk containing the byte, among the
     * array of the controller's disks.
     *
     * @return The index.
     **/

    public int getDiskIndex()
    {
        return mDiskIndex;
    }

    /**
     * Returns the index of the block containing the byte, among the
     * blocks of the containing disk.
     *
     * @return The index.
     **/

    public int getBlockIndex()
    {
        return mBlockIndex;
    }

    /**
     * Returns the block data offset of the byte.
     *
     * @return The data offset.
     **/

    public int getBlockOffset()
    {
        return mBlockOffset;
    }

    /**
     * Returns the disk containing the byte.
     *
     * @return The disk.
     **/

    public Disk getDisk()
    {
        return getDisks()[getDiskIndex()];
    }

    /**
     * Returns the block containing the byte.
     *
     * @return The block.
     *
     * @throws DiskException Thrown iff a disk error occurs.
     **/

    public Block getBlock()
        throws DiskException
    {
        return getDisk().getBlock(getBlockIndex());
    }

    /**
     * Returns the byte data.
     *
     * @return The data.
     *
     * @throws DiskException Thrown iff a disk error occurs.
     **/

    public byte getData()
        throws DiskException
    {
        return getBlock().getData(getBlockOffset());
    }

    /**
     * Sets the byte to the given value.
     *
     * @param data The value.
     *
     * @throws DiskException Thrown iff a disk error occurs.
     **/

    public void setData(byte data)
        throws DiskException
    {
        getBlock().setData(getBlockOffset(),data);
    }
}
