package csci4534.controller;

/**
 * A disk controller. The controller manages a set of <EM>n</EM>
 * identical disks, meaning that all disks have the same number
 * <EM>n<SUB>b</SUB></EM> of blocks and block size
 * <EM>c<SUB>b</SUB></EM> bytes. Initially (during formatting), all
 * disks can be assumed to be good, and all their blocks are
 * accessible and zeroed out. Over time, however, the disks and their
 * blocks may get damaged.
 *
 * <P>
 *
 * The <EM>total capacity</EM> that the disks can accomodate is of
 * course
 * <EM>c<SUB>t</SUB></EM>=<EM>n</EM>*<EM>n<SUB>b</SUB></EM>*<EM>c<SUB>b</SUB></EM>
 * bytes. However, when the disks get formatted, the user may specify
 * that only <EM>c<SUB>a</SUB></EM> bytes will be accessed via {@link
 * #read(int)} and {@link #write(int,byte)}; this is the <EM>actual
 * capacity</EM>. The controller may use the remaining
 * <EM>c<SUB>t</SUB></EM>-<EM>c<SUB>a</SUB></EM> bytes however it
 * wishes. A common use of the remaining storage is to provide
 * reliability, either as spare blocks on each disk or as metadata in
 * a RAID (i.e. storing duplicates of blocks or computed
 * parity). Different controllers have different disk management
 * policies, and so may impose different constraints between
 * <EM>c<SUB>a</SUB></EM> and <EM>c<SUB>t</SUB></EM>. For example, all
 * controllers require
 * <EM>c<SUB>a</SUB></EM>&lt;=<EM>c<SUB>t</SUB></EM>; but a RAID 1
 * controller requires
 * <EM>c<SUB>a</SUB></EM>&lt;=floor(<EM>n</EM>/2)*<EM>n<SUB>b</SUB></EM>*<EM>c<SUB>b</SUB></EM>
 * because disks are mirrored for reliability.
 *
 * <P>
 *
 * A controller has an NVRAM chip, on which it can store a constant
 * amount of data for each disk. In other words, implementations of
 * this interface can have any number of instance variables provided
 * each consumes at most <EM>O(n)</EM> bytes. This limit includes
 * direct consumption, e.g. a variable like an array, as well as
 * indirect consumption, e.g. a variable which is an instance of a
 * class, which in turn has instance variables, and so on
 * recursively. Another way to state this constraint is that two
 * controllers
 *
 * <UL>
 * <LI> of the same type, and
 * <LI> which differ only in <EM>n<SUB>b</SUB></EM>,
 * <EM>c<SUB>b</SUB></EM>, and/or <EM>c<SUB>a</SUB></EM>,
 * </UL>
 *
 * will use exactly the same amount of memory in their instance
 * variables (directly or indirectly).
 *
 * <P>
 *
 * A single call to {@link #format(Disk[],int)} preceeds all other
 * calls to this interface. This is where the controller initializes
 * its NVRAM. It is also at this point that the controller throws an
 * exception if its constraints on <EM>c<SUB>a</SUB></EM> and
 * <EM>c<SUB>t</SUB></EM> are not satisfied; in this case, no further
 * calls are made into the controller (as formatting failed, the disks
 * are unusable).
 *
 * <P>
 *
 * If the disks get formatted successfully, the user then reads and
 * writes data. Specifically, the user treats the controller as a
 * byte-addressable device, retrieving or setting one byte at a
 * time. At this point, the fact that the controller manages multiple
 * disks --- each with multiple blocks --- is completely hidden from
 * the user (until errors occur, at which point instances of {@link
 * DiskException} get thrown, revealing the use of disks to store the
 * data). The user's view of the controller is like that of a large
 * byte array which can read/write one byte at a time, and where each
 * byte is accessible via its index (called the
 * <EM>offset</EM>). Naturally, the offset should be in the range
 * [0,<EM>c<SUB>a</SUB></EM>-1]. This is, of course, an
 * abstraction. Given the above constraint on the controller's
 * instance variables, the controller cannot allocate such an array
 * (directly or indirectly); instead, all data is stored on the disks.
 *
 * <P>
 *
 * If an underlying disk gets completely or partially damaged, this
 * will not necessarily cause a read/write failure at the level of the
 * controller. For example, if a primary disk fails while the
 * corresponding mirror disk is good, a RAID 1 controller can continue
 * to read/write data on the mirror disk; as a result, controller
 * read/write operations will continue to succeed. Of course, whether
 * disk problems affect the controller's ability to carry out
 * operations depends on the controller's disk management policy. For
 * example, by contrast to a RAID controller, a simple controller
 * supporting no spares or redundant storage will immediately report
 * failures as soon as disk problems occur.
 *
 * <P>
 *
 * Some disk failures are pretty straight-forward: the controller
 * either completely hides them (see the RAID controller above) or it
 * fails also (see the simple controller above). Many disk failures are
 * not straight-forward. Consider for example, a controller that
 * supports spare blocks or disks:
 *
 * <UL>
 *
 * <LI> If we attempt to read data from an area that just got damaged,
 * then the attempt will fail. The controller will transparently
 * replace the damaged item with a spare (assuming one is available),
 * but that will only help future operations. As far as this
 * particular read right now was concerned, it failed; yet future
 * operations on the same offset will succeed.
 *
 * <P>
 *
 * <LI> Similarly, consider a write operation to an area that just got
 * damaged. After the spare is put in place, the operation can in fact
 * succeed. That doesn't mean that there was no data loss: whatever
 * data was on the damaged area is unrecoverable. So, even though
 * future reads will succeed, in general they will not return the same
 * data the user had stored prior to the failure.
 *
 * </UL>
 *
 * In the above situations, does the controller report success or
 * failure? We adopt the following convention: if the disk failure
 * cannot be <STRONG>completely</STRONG> hidden, we report a
 * controller failure. More precisely, we report a disk failure
 *
 * <UL>
 * <LI> as soon as we detect it,
 * <LI> provided this failure caused true data loss to the controller user,
 * <LI> whether or not that loss directly impacts the requested operation.
 * </UL>
 *
 * So, in both examples above, {@link #read(int)} or {@link
 * #write(int,byte)} rethrow the {@link DiskException} that lead to
 * their failure. That said, before we return from either method, we
 * take all corrective steps we can, such as installing the
 * spare. Clarifications:
 *
 * <UL>
 *
 * <LI> In the second example above, we install the spare, but the
 * write does not record the new data on the spare; this is because
 * the user will see a failure anyway so they will most likely attempt
 * a re-write.
 *
 * <P>
 *
 * <LI> Reduced reliability is not data loss. On a RAID system, if a
 * disk completely fails but its mirror is good and has all blocks
 * accessible, then no data loss has occured from the perspective of
 * the controller user (though future data loss is now more
 * likely). So we report no failure. In real systems, a system log
 * entry is generated to inform the user of reliability loss, but no
 * other action is taken.
 *
 * </UL>
 *
 * <P>
 *
 * As stated above, some controllers can take corrective steps. For
 * example,
 *
 * <UL>
 *
 * <LI> A spare controller can install a spare. What if there are no
 * spares available?
 *
 * <P>
 *
 * <LI> Or, what if, as we search for a spare, we discover additional
 * damaged areas, yet eventually we do find a spare?
 *
 * <P>
 *
 * <LI> A RAID 1 controller uses a mirror when the primary disk
 * fails. What if the mirror also fails at the same time as the
 * primary disk does?
 * 
 * </UL>
 *
 * In all of the above cases, the corrective action encounters more
 * disk problems. In the end, the controller may or may not recover
 * from these additional failures. Regardless, the earlier convention
 * still applies: if the disk failure cannot be completely hidden, we
 * report a controller failure. Moreover, we report only the first
 * disk failure, i.e. the one that lead us into taking corrective
 * action; in other words, we rethrow the first {@link DiskException}
 * we encounter. Note that, by <EM>first</EM> we don't mean that we go
 * back in time over the controller's history. For example, if the
 * primary disk fails, and after many more operations the mirror
 * fails, we don't report the original primary disk failure. By
 * <EM>first</EM> we mean the earliest failure within the current
 * context of the operation we are performing.
 * 
 * @author Toli Lerios
 **/

public interface Controller
{

    // PUBLIC INTERFACE.

    /**
     * Formats the given set of disks to accomodate the desired actual
     * capacity. Also, the controller initializes its NVRAM.
     *
     * @param disks The <EM>n</EM> identical disks, which are all
     * guaranteed to be good and have no inaccessible blocks.
     * @param actualCapacity The actual capacity <EM>c<SUB>a</SUB></EM>.
     *
     * @throws InsufficientSpaceException Thrown iff the controller's
     * disk management policy cannot accomodate the provided actual
     * capacity <EM>c<SUB>a</SUB></EM> within the total capacity
     * <EM>c<SUB>t</SUB></EM>.
     **/

    public void format(Disk[] disks,
                       int actualCapacity)
        throws InsufficientSpaceException;

    /**
     * Returns the controller data at the given offset. If any disk
     * failures occur, the operation succeeds only if those failures
     * can be completely hidden. If any disk failures occur,
     * corrective action (if any) is taken before this method
     * completes (successfully or not).
     *
     * @param offset The offset.
     *
     * @return The data.
     *
     * @throws DiskException Thrown iff data loss occured. In case of
     * multiple failures during corrective action, only the first
     * failure is reported.
     **/

    public byte read(int offset)
        throws DiskException;

    /**
     * Sets the controller data at the given offset to the given data
     * value. If any disk failures occur, the operation succeeds only
     * if those failures can be completely hidden. If any disk
     * failures occur, corrective action (if any) is taken before this
     * method completes (successfully or not).
     *
     * @param offset The offset.
     * @param data The data value.
     *
     * @throws DiskException Thrown iff data loss occured. In case of
     * multiple failures during corrective action, only the first
     * failure is reported.
     **/

    public void write(int offset,
                      byte data)
        throws DiskException;
}
