package csci4534.controller;

/**
 * A foundation class for controller implementations. This class
 * retains the controller disks and actual capacity via {@link
 * #format(Disk[],int)}, and offers convenient shorthands. The {@link
 * #format(Disk[],int)} implementation of this class should be called
 * by all subclasses before they invoke any other method.
 * 
 * @author Toli Lerios
 **/

abstract public class ControllerBase
    implements Controller
{

    // INSTANCE DATA.

    private Disk[] mDisks;
    private int mActualCapacity;


    // PUBLIC INTERFACE.

    /**
     * Returns the controller's disks.
     *
     * @return The disks.
     **/

    public Disk[] getDisks()
    {
        return mDisks;
    }

    /**
     * Returns the number <EM>n<SUB>b</SUB></EM> of blocks on each of
     * the controller's disks.
     *
     * @return The number of blocks.
     **/

    public int getBlockCount()
    {
        return getDisks()[0].getBlockCount();
    }

    /**
     * Returns the block size <EM>c<SUB>b</SUB></EM> of the
     * controller's disks.
     *
     * @return The block size, in bytes.
     **/

    public int getBlockSize()
    {
        return getDisks()[0].getBlockSize();
    }

    /**
     * Returns the total number of bytes that can be stored on each of
     * the controller's disks. This is
     * <EM>n<SUB>b</SUB></EM>*<EM>c<SUB>b</SUB></EM>.
     *
     * @return The number of bytes.
     **/

    public int getTotalDiskCapacity()
    {
        return getBlockCount()*getBlockSize();
    }

    /**
     * Returns the total capacity <EM>c<SUB>t</SUB></EM> of the
     * controller's disks.
     *
     * @return The total capacity.
     **/

    public int getTotalCapacity()
    {
        return getDisks().length*getTotalDiskCapacity();
    }

    /**
     * Returns the actual capacity <EM>c<SUB>a</SUB></EM> of the
     * controller.
     *
     * @return The actual capacity.
     **/

    public int getActualCapacity()
    {
        return mActualCapacity;
    }

    /**
     * Throws an {@link InsufficientSpaceException} with the given
     * constraint, and the capacities associated with the controller.
     *
     * @param constraint The constraint.
     *
     * @throws InsufficientSpaceException Always throws.
     **/

    public void throwInsufficientSpace(String constraint)
        throws InsufficientSpaceException
    {
        throw new InsufficientSpaceException
            (getActualCapacity(),getTotalCapacity(),constraint);
    }


    // Controller.

    public void format(Disk[] disks,
                       int actualCapacity)
        throws InsufficientSpaceException
    {
        mDisks=disks;
        mActualCapacity=actualCapacity;
    }
}
