package csci4534.controller;

/**
 * A disk. It has two parts: 
 *
 * <OL>
 *
 * <LI> An NVRAM chip, on which the disk metadata is stored: these
 * comprise the disk name (an arbitrary string), the number
 * <EM>n<SUB>b</SUB></EM> of disk blocks, and the size
 * <EM>c<SUB>b</SUB></EM> of each block (in bytes).
 *
 * <LI> The disk platters, on which the disk data is stored. Disk data
 * is manipulated as complete blocks, retrieved from the disk via a
 * block index. Naturally, the block index should be in the range
 * [0,<EM>n<SUB>b</SUB></EM>-1]. Once a block is retrieved from the
 * disk, its contents can be read/written via the methods of {@link
 * Block}.
 *
 * </OL>
 *
 * The disk starts in working state (<EM>good</EM>) whereby all blocks
 * are accessible. The disk platters may get partially damaged, at
 * which point some blocks become inaccessible; a disk is still
 * considered good in this case. The disk platters may also become
 * completely damaged (e.g. if an internal chip burns out), at which
 * point no block is accessible; such a disk is no longer good. Note
 * that if all blocks of a good disk become inaccessible one by one,
 * the disk is still considered good; in other words, block
 * accessibility does not affect disk good-ness (though disk good-ness
 * affects block accessibility). Regardless of the state of the disk
 * and its blocks, the NVRAM information is always available.
 *
 * <P>
 *
 * A good disk also supports block remapping. This means that, if
 * block <EM>i</EM> becomes inaccessible, the disk user can request
 * that block <EM>i</EM> be swapped with any other block
 * <EM>b</EM>. After this swap, a request for block <EM>i</EM> will in
 * fact return the block previously accessible under the index
 * <EM>b</EM>; while a request for block <EM>b</EM> will fail because
 * block <EM>b</EM> now maps to the inaccessible block.
 *
 * @author Toli Lerios
 **/

public class Disk
{

    // INSTANCE DATA.

    private String mName;
    private int mBlockCount;
    private int mBlockSize;
    private Block[] mBlocks;


    // PRIVATE HELPERS.

    /**
     * Returns the given disk block, assuming the disk is good and the
     * block is accessible.
     *
     * @param index The block index.
     *
     * @return The block.
     **/

    private Block getBlockUnsafe(int index)
    {
        return mBlocks[index];
    }

    /**
     * Sets the given disk block, assuming the disk is good and the
     * block is accessible.
     *
     * @param index The block index.
     * @param block The new block.
     **/

    private void setBlockUnsafe(int index,
                                Block block)
    {
        mBlocks[index]=block;
    }

    /**
     * Checks if the given disk block is accessible, assuming the disk
     * is good.
     *
     * @param index The block index.
     *
     * @return True iff it is accessible.
     **/

    private boolean isGoodBlockUnsafe(int index)
    {
        return (getBlockUnsafe(index)!=null);
    }

    /**
     * Asserts that the given disk block is accessible, assuming the
     * disk is good.
     *
     * @param index The block index.
     *
     * @throws BadBlockException Thrown iff the block is inaccessible.
     **/

    private void assertGoodBlock(int index)
        throws BadBlockException
    {
        if (!isGoodBlockUnsafe(index)) {
            throw new BadBlockException(this,index);
        }
    }

    /**
     * Asserts that the disk is good.
     *
     * @throws BadDiskException Thrown iff the disk is no longer good.
     **/

    private void assertGood()
        throws BadDiskException
    {
        if (!isGood()) {
            throw new BadDiskException(this);
        }
    }


    // CONSTRUCTORS.

    /**
     * Creates a new disk. The new disk's data is all zeroed out.
     *
     * @param name The disk name.
     * @param blockCount The number <EM>n<SUB>b</SUB></EM> of blocks.
     * @param blockSize The size <EM>c<SUB>b</SUB></EM> of each block.
     **/

    public Disk(String name,
                int blockCount,
                int blockSize)
    {
        mName=name;
        mBlockCount=blockCount;
        mBlockSize=blockSize;
        mBlocks=new Block[getBlockCount()];
        for (int i=0;i<getBlockCount();i++) {
            setBlockUnsafe(i,new Block(getBlockSize()));
        }
    }


    // PUBLIC INTERFACE.

    /**
     * Returns the disk name.
     *
     * @return The name.
     **/

    public String getName()
    {
        return mName;
    }

    /**
     * Returns the block size <EM>c<SUB>b</SUB></EM>.
     *
     * @return The size.
     **/

    public int getBlockSize()
    {
        return mBlockSize;
    }

    /**
     * Returns the number <EM>n<SUB>b</SUB></EM> of blocks.
     *
     * @return The number of blocks.
     **/

    public int getBlockCount()
    {
        return mBlockCount;
    }

    /**
     * Checks if the disk is good (though some blocks may be
     * inaccessible).
     *
     * @return True iff it is good.
     **/

    public boolean isGood()
    {
        return (mBlocks!=null);
    }

    /**
     * Checks if the given disk block is accessible.
     *
     * @param index The block index.
     *
     * @return True iff it is accessible.
     *
     * @throws BadDiskException Thrown iff the disk is no longer good.
     **/

    public boolean isGoodBlock(int index)
        throws BadDiskException
    {
        assertGood();
        return isGoodBlockUnsafe(index);
    }

    /**
     * Returns the given disk block.
     *
     * @param index The block index.
     *
     * @return The block.
     *
     * @throws BadDiskException Thrown iff the disk is no longer good.
     * @throws BadBlockException Thrown iff the block is inaccessible.
     **/

    public Block getBlock(int index)
        throws BadDiskException,
               BadBlockException
    {
        assertGood();
        assertGoodBlock(index);
        return getBlockUnsafe(index);
    }

    /**
     * Completely damages the disk, thus making it no longer a good
     * one.
     **/

    public void damage()
    {
        mBlocks=null;
    }

    /**
     * Damages the given disk block.
     *
     * @param index The block index.
     *
     * @throws BadDiskException Thrown iff the disk is no longer good.
     **/

    public void damageBlock(int index)
        throws BadDiskException
    {
        assertGood();
        setBlockUnsafe(index,null);
    }

    /**
     * Remaps the given disk block to a new index.
     *
     * @param oldIndex The original block index.
     * @param newIndex The new block index.
     *
     * @throws BadDiskException Thrown iff the disk is no longer good.
     **/

    public void remapBlock(int oldIndex,
                           int newIndex)
        throws BadDiskException
    {
        assertGood();
        Block tmp=getBlockUnsafe(newIndex);
        setBlockUnsafe(newIndex,getBlockUnsafe(oldIndex));
        setBlockUnsafe(oldIndex,tmp);
    }


    // Object.

    public String toString()
    {
        StringBuffer buffer=new StringBuffer();
        buffer.append(getName());
        buffer.append(':');
        if (isGood()) {
            buffer.append('[');
            for (int i=0;i<getBlockCount();i++) {
                if (i>0) {
                    buffer.append(';');
                }
                if (isGoodBlockUnsafe(i)) {
                    buffer.append(getBlockUnsafe(i));
                } else {
                    buffer.append("bad");
                }
            }
            buffer.append(']');
        } else {
            buffer.append("bad");
        }
        return buffer.toString();
    }
}
