package csci4534.controller;

/**
 * A disk controller with spare blocks and striping (but no redundant
 * storage). Blocks which are not used to store controller data are
 * used as spares. The number of spares per disk is not the same for
 * all disks. For example, for <EM>c<SUB>a</SUB></EM>=5, <EM>n</EM>=4,
 * <EM>c<SUB>b</SUB></EM>=1, we use 2 blocks on the first disk for
 * controller data, but only 1 block on each of the other three disks;
 * conversely, the last three disks have one more spare block than the
 * first one.
 *
 * <P>
 *
 * If a block becomes inaccessible, then it always results into a
 * controller failure, but before the error is reported, a spare is
 * installed: in other words, the damaged block is replaced with one
 * of the spare blocks (if any, on the same disk). If that spare block
 * is accessible (which is something we do not test when we install
 * it), then later calls to access its bytes will succeed and so the
 * offset that caused the earlier failure will no longer cause a
 * failure. If a later attempt to access the newly installed spare
 * block fails because it is also inaccessible, we try installing
 * another spare, and so on until we run out of spares. At this point,
 * or if the disk becomes no longer good, then those controller
 * failures become permanent (i.e. byte requests with offsets mapped
 * to the damaged areas will never succeed again).
 *
 * @author Toli Lerios
 **/

public class SpareController
    extends ControllerBase
{

    // INSTANCE DATA.

    private int[] mSpareBlocks;


    // PRIVATE HELPERS.

    /**
     * Attempts to install a spare block on the disk containing the
     * given inaccessible block. This attempt may fail transparently
     * if no spares are available of if the disk is no longer
     * good. Note that the accessibility of the newly installed spare
     * is not tested.
     *
     * @param position The byte position pointing to the inaccessible
     * block (and its associated disk).
     **/

    private void installSpare(BytePosition position)
    {
        if (mSpareBlocks[position.getDiskIndex()]>0) {
            try {
                position.getDisk().remapBlock
                    (position.getDisk().getBlockCount()-
                     mSpareBlocks[position.getDiskIndex()],
                     position.getBlockIndex());
            } catch (BadDiskException ex) {
                // If the disk is bad, there's nothing we can do.
                return;
            }
            mSpareBlocks[position.getDiskIndex()]--;
        }
    }


    // ControllerBase.

    public void format(Disk[] disks,
                       int actualCapacity)
        throws InsufficientSpaceException
    {
        super.format(disks,actualCapacity);
        // The total number of blocks needed to store the controller
        // bytes.
        int actualBlocks=Utilities.divUp(actualCapacity,getBlockSize());
        // The minimum number of spare blocks per disk.
        int spareCount=getBlockCount()-
            Utilities.divUp(actualBlocks,getDisks().length);
        if (spareCount<1) {
            throwInsufficientSpace("Actual<=Total-(1 spare block per disk)");
        }
        // The first disk to have an extra spare block.
        int firstExtra=(actualBlocks%getDisks().length);
        if (firstExtra==0) {
            firstExtra=getDisks().length;
        }
        // Setup initial counts of available spare blocks per disk.
        mSpareBlocks=new int[getDisks().length];
        for (int i=0;i<firstExtra;i++) {
            mSpareBlocks[i]=spareCount;
        }
        spareCount++;
        for (int i=firstExtra;i<getDisks().length;i++) {
            mSpareBlocks[i]=spareCount;
        }
    }

    public byte read(int offset)
        throws DiskException
    {
        BytePosition position=
            new StripedBytePosition(offset,getDisks().length,getDisks());
        try {
            return position.getData();
        } catch (DiskException ex) {
            // Attempt to install a spare before reporting failure.
            installSpare(position);
            throw ex;
        }
    }

    public void write(int offset,
                      byte data)
        throws DiskException
    {
        BytePosition position=
            new StripedBytePosition(offset,getDisks().length,getDisks());
        try {
            position.setData(data);
        } catch (DiskException ex) {
            // Attempt to install a spare before reporting failure.
            installSpare(position);
            throw ex;
        }
    }
}


